# SPDX-License-Identifier: MIT
"""CSV Event Listing."""

import logging
import os
import csv

import metarace
from metarace import strops

LOG = logging.getLogger('metarace.eventdb')
LOG.setLevel(logging.DEBUG)

# Note: These are for the trackmeet module, roadmeet re-defines race types
defracetypes = [
    'sprint',
    'keirin',
    'flying 200',
    'flying lap',
    'indiv tt',
    'indiv pursuit',
    'pursuit race',
    'points',
    'madison',
    'omnium',
    'classification',
    'hour',
    'competition',
    'break',
    'sprint round',
    'sprint final',
    'scratch',
    'motorpace',
    'handicap',
    'elimination',
    'race',
]

# default event values (if not empty string)
EVENT_DEFAULTS = {
    'evid': None,  # empty not allowed
    'resu': True,
    'inde': False,
    'prin': False,
    'dirt': False,
    'plac': None,
    'laps': None,
}

# event column heading and key mappings
EVENT_COLUMNS = {
    'evid': "EvID",
    'refe': "Reference Number",
    'pref': "Prefix",
    'info': "Information",
    'seri': "Series",
    'type': "Type Handler",
    'star': "Starters",
    'depe': "Depends On Events",
    'resu': "Result Include?",
    'inde': "Index Include?",
    'prin': "Printed Program Include?",
    'plac': "Placeholder Count",
    'sess': "Session",
    'laps': "Laps Count",
    'dist': "Distance String",
    'prog': "Progression Rules String",
    'reco': "Record String",
    'dirt': "Dirty?",
    'evov': "EVOverride"
}

# for any non-strings, types as listed
EVENT_COLUMN_CONVERTERS = {
    'resu': strops.confopt_bool,
    'inde': strops.confopt_bool,
    'prin': strops.confopt_bool,
    'dirt': strops.confopt_bool,
    'plac': strops.confopt_posint,
    'laps': strops.confopt_posint,
}

DEFAULT_COLUMN_ORDER = [
    'evid', 'refe', 'pref', 'info', 'seri', 'type', 'star', 'depe', 'resu',
    'inde', 'prin', 'plac', 'sess', 'laps', 'dist', 'prog', 'evov', 'reco',
    'dirt'
]


def colkey(colstr=''):
    """Convert a coumn header string to a colkey."""
    return colstr[0:4].lower()


def get_header(cols=DEFAULT_COLUMN_ORDER):
    """Return a row of header strings for the provided cols."""
    return [EVENT_COLUMNS[c] for c in cols]


class event(object):
    """CSV-backed event listing."""

    def get_row(self, coldump=DEFAULT_COLUMN_ORDER):
        """Return a row ready to export."""
        return [str(self[c]) for c in coldump]

    def event_info(self):
        """Return a concatenated and stripped event information string."""
        return ' '.join([self['pref'], self['info']]).strip()

    def event_type(self):
        """Return event type string."""
        return self['type'].capitalize()

    def set_notify(self, callback=None):
        """Set or clear the notify callback for the event."""
        if callback is not None:
            self.__notify = callback
        else:
            self.__notify = self.__def_notify

    def get_value(self, key):
        """Alternate value fetch."""
        return self.__getitem__(key)

    def set_value(self, key, value):
        """Update a value without triggering notify."""
        key = colkey(key)
        self.__store[key] = value

    def notify(self):
        """Forced notify."""
        self.__notify(self.__store['evid'])

    def __init__(self, evid=None, notify=None, cols={}):
        self.__store = dict(cols)
        self.__notify = self.__def_notify
        if 'evid' not in self.__store:
            self.__store['evid'] = evid
        if notify is not None:
            self.__notify = notify

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        """Use a default value id, but don't save it."""
        key = colkey(key)
        if key in self.__store:
            return self.__store[key]
        elif key in EVENT_DEFAULTS:
            return EVENT_DEFAULTS[key]
        else:
            return ''

    def __setitem__(self, key, value):
        key = colkey(key)
        self.__store[key] = value
        self.__notify(self.__store['evid'])

    def __delitem__(self, key):
        key = colkey(key)
        del (self.__store[key])
        self.__notify(self.__store['evid'])

    def __iter__(self):
        return iter(self.__store.keys())

    def iterkeys(self):
        return iter(self.__store.keys())

    def __contains__(self, item):
        key = colkey(item)
        return key in self.__store

    def __def_notify(self, data=None):
        pass


class eventdb(object):
    """Event database."""

    def add_empty(self, evno=None):
        """Add a new empty row to the event model."""
        if evno is None:
            evno = self.nextevno()
        nev = event(evid=evno, notify=self.__notify)
        self.__store[evno] = nev
        self.__index.append(evno)
        self.__notify(None)
        LOG.debug('Added empty event %r', evno)
        return nev

    def clear(self):
        """Clear event model."""
        self.__index = []
        self.__store = {}
        self.__notify(None)
        LOG.debug('Event model cleared')

    def change_evno(self, oldevent, newevent):
        """Attempt to change the event id."""
        if oldevent not in self:
            LOG.error('Change event %r not found', oldevent)
            return False

        if newevent in self:
            LOG.error('New event %r already exists', newevent)
            return False

        oktochg = True
        if self.__evno_change_cb is not None:
            oktochg = self.__evno_change_cb(oldevent, newevent)
        if oktochg:
            ref = self.__store[oldevent]
            ref.set_value('evid', newevent)
            cnt = 0
            for j in self.__index:
                if j == oldevent:
                    break
                cnt += 1
            if cnt < len(self.__index):
                self.__index[cnt] = newevent
            del (self.__store[oldevent])
            self.__store[newevent] = ref
            LOG.info('Updated event %r to %r', oldevent, newevent)
            return True
        return False

    def add_event(self, newevent):
        """Append newevent to model."""
        eid = newevent['evid']
        if eid is None:
            eid = self.nextevno()
        elif not isinstance(eid, str):
            LOG.debug('Converted %r to event id: %r', eid, str(eid))
            eid = str(eid)
        evno = eid
        while evno in self.__index:
            evno = u'-'.join((eid, strops.randstr()))
            LOG.info('Duplicate evid %r changed to %r', eid, evno)
        newevent['evid'] = evno
        LOG.debug('Add new event with id=%r', evno)
        newevent.set_notify(self.__notify)
        self.__store[evno] = newevent
        self.__index.append(evno)

    def __loadrow(self, r, colspec):
        nev = event()
        for i in range(0, len(colspec)):
            if len(r) > i:  # column data in row
                val = r[i].translate(strops.PRINT_UTRANS)
                key = colspec[i]
                if key in EVENT_COLUMN_CONVERTERS:
                    val = EVENT_COLUMN_CONVERTERS[key](val)
                nev[key] = val
        if not nev['evid']:
            evno = self.nextevno()
            LOG.info('Event without id assigned %r', evno)
            nev['evid'] = evno
        self.add_event(nev)

    def load(self, csvfile=None):
        """Load events from supplied CSV file."""
        if not os.path.isfile(csvfile):
            LOG.debug('Events file %r not found', csvfile)
            return
        LOG.debug('Loading events from %r', csvfile)
        with open(csvfile, 'r', encoding='utf-8') as f:
            cr = csv.reader(f)
            incols = None  # no header
            for r in cr:
                if len(r) > 0:  # got a data row
                    if incols is not None:  # already got col header
                        self.__loadrow(r, incols)
                    else:
                        # determine input column structure
                        if colkey(r[0]) in EVENT_COLUMNS:
                            incols = []
                            for col in r:
                                incols.append(colkey(col))
                        else:
                            incols = DEFAULT_COLUMN_ORDER  # assume full
                            self.__loadrow(r, incols)
        self.__notify(None)

    def save(self, csvfile=None):
        """Save current model content to CSV file."""
        if len(self.__index) != len(self.__store):
            LOG.error('Index out of sync with model, dumping whole model')
            self.__index = [a for a in self.__store]

        LOG.debug('Saving events to %r', csvfile)
        with metarace.savefile(csvfile) as f:
            cr = csv.writer(f, quoting=csv.QUOTE_ALL)
            cr.writerow(get_header(self.include_cols))
            for r in self:
                cr.writerow(r.get_row())

    def nextevno(self):
        """Try and return a new event number string."""
        lmax = 1
        for r in self.__index:
            if r.isdigit() and int(r) >= lmax:
                lmax = int(r) + 1
        return str(lmax)

    def set_evno_change_cb(self, cb, data=None):
        """Set the event no change callback."""
        self.__evno_change_cb = cb

    def getfirst(self):
        """Return the first event in the db."""
        ret = None
        if len(self.__index) > 0:
            ret = self[self.__index[0]]
        return ret

    def getnextrow(self, ref, scroll=True):
        """Return reference to the row one after current selection."""
        ret = None
        if ref is not None:
            path = self.__index.index(ref['evid']) + 1
            if path >= 0 and path < len(self.__index):
                ret = self[self.__index[path]]  # check reference
        return ret

    def getprevrow(self, ref, scroll=True):
        """Return reference to the row one after current selection."""
        ret = None
        if ref is not None:
            path = self.__index.index(ref['evid']) - 1
            if path >= 0 and path < len(self.__index):
                ret = self[self.__index[path]]  # check reference
        return ret

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        return self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = value  # no change to key
        self.__notify(key)

    def __delitem__(self, key):
        self.__index.remove(key)
        del (self.__store[key])

    def __iter__(self):
        for r in self.__index:
            yield self.__store[r]

    def iterkeys(self):
        return self.__index.__iter__()

    def __contains__(self, item):
        return item in self.__store

    def set_notify(self, cb=None):
        """Set the data change notification callback."""
        if cb is None:
            cb = self.__defnotify
        self.__notify = cb
        for ev in self:
            ev.set_notify(cb)

    def __def_notify(self, data=None):
        """Handle changes in db."""
        pass

    def __init__(self, racetypes=None):
        """Constructor for the event db."""
        self.__index = []
        self.__store = {}
        self.__notify = self.__def_notify
        self.__evno_change_cb = None

        self.include_cols = DEFAULT_COLUMN_ORDER
        if racetypes is not None:
            self.racetypes = racetypes
        else:
            self.racetypes = defracetypes



# SPDX-License-Identifier: MIT
"""CSV-backed Competitor and Category Data"""

import logging
import os
import csv
from unicodedata import normalize
import grapheme

import metarace
from metarace import strops

_log = logging.getLogger('metarace.riderdb')
_log.setLevel(logging.DEBUG)

# default values when columns omitted
_RIDER_DEFAULTS = {}

# Rider column headings
_RIDER_COLUMNS = {
    'no': 'Rider No',
    'series': 'Series',
    'first': 'First Name',
    'last': 'Last Name',
    'org': 'Organisation',
    'cat': 'Categories',
    'nat': 'Nationality',
    'ref': 'Refid/Transponder',
    'uci': 'UCI ID',
    'dob': 'DoB',
    'sex': 'Sex',
    'note': 'Notes',
}

# Category column headings
_CATEGORY_COLUMNS = {
    'no': 'ID',
    'series': 'Series',
    'first': 'Title',
    'last': 'Subtitle',
    'org': 'Lap Prefix',
    'cat': 'Target Laps',
    'nat': 'Nationality',
    'uci': 'Start Offset',
    'ref': 'Distance',
    'note': 'Footer',
    'dob': 'DoB',
    'sex': 'Sex',
}

# legacy csv file ordering
_DEFAULT_COLUMN_ORDER = [
    'no', 'first', 'last', 'org', 'cat', 'series', 'ref', 'uci', 'dob', 'nat',
    'sex', 'note'
]

# Alternative column header strings lookup
_ALT_COLUMNS = {
    'id': 'no',
    'rid': 'no',
    'num': 'no',
    'cid': 'no',
    'cod': 'no',
    'bib': 'no',
    'ind': 'no',
    'no.': 'no',
    'ser': 'series',
    'typ': 'series',
    'fir': 'first',
    'nam': 'first',
    'tit': 'first',
    'las': 'last',
    'sub': 'last',
    'sur': 'last',
    'clu': 'org',
    'tea': 'org',
    'lap': 'org',
    'tar': 'cat',
    'lic': 'uci',
    'sta': 'uci',
    'rfi': 'ref',
    'tra': 'ref',
    'dis': 'ref',
    'gen': 'sex',
    'dat': 'dob',
    'not': 'note',
    'foo': 'note',
}


def colkey(colstr=''):
    """Convert a column header string to a colkey."""
    col = colstr[0:3].lower()
    if col in _ALT_COLUMNS:
        col = _ALT_COLUMNS[col]
    return col


def get_header(cols=_DEFAULT_COLUMN_ORDER, hdrs=_RIDER_COLUMNS):
    """Return a row of header strings for the provided cols."""
    return (hdrs[colkey(c)] for c in cols)


def cellnorm(unistr):
    """Normalise supplied string, then return a version with printing chars."""
    return normalize('NFC', unistr).translate(strops.PRINT_UTRANS)


class rider():
    """Rider handle."""

    def get_id(self):
        """Return this rider's unique id"""
        return (self.__store['no'].lower(), self.__store['series'].lower())

    def get_key(self):
        """Return a sorting key for this rider number"""
        return strops.bibstr_key(self.__store['no'])

    def primary_cat(self):
        """Return rider's primary category"""
        ret = ''
        cv = self['cat'].split()
        if cv:
            ret = cv[0].upper()
        return ret

    def get_cats(self):
        """Return a list of categories for this rider"""
        return (c.upper() for c in self['cat'].split())

    def get_row(self, coldump=_DEFAULT_COLUMN_ORDER):
        """Return a row ready to export."""
        return (str(self[c]) for c in coldump)

    def set_notify(self, callback=None):
        """Set or clear the notify callback for the event."""
        if callback is not None:
            self.__notify = callback
        else:
            self.__notify = self.__def_notify

    def get_value(self, key):
        """Alternate value fetch."""
        return self.__getitem__(key)

    def set_value(self, key, value):
        """Update a value without triggering notify."""
        key = colkey(key)
        self.__store[key] = value

    def notify(self):
        """Forced notify."""
        self.__notify(self.get_id())

    def __init__(self, cols={}, no='', series='', notify=None):
        self.__strcache = {}
        self.__store = dict(cols)
        self.__notify = self.__def_notify
        if 'no' not in self.__store:
            self.__store['no'] = no
        if 'series' not in self.__store:
            self.__store['series'] = series
        if notify is not None:
            self.__notify = notify

    def fitname(self, width, trunc=False):
        """Return a truncated name string of width or less graphemes"""
        ret = None
        nkey = ('fn', width, trunc)
        if nkey not in self.__strcache:
            fn = self['first'].strip().title()
            fl = grapheme.length(fn)
            ln = self['last'].strip().upper()
            ll = grapheme.length(ln)
            if fl + ll >= width:
                lshrt = ln.split('-')[-1].strip()
                lsl = grapheme.length(lshrt)
                ln = lshrt
                if fl + lsl >= width:
                    if fl > 2:
                        fn = grapheme.slice(fn, end=1) + '.'
            ret = ' '.join((fn, ln))
            if trunc and grapheme.length(ret) > width:
                if width > 4:
                    ret = grapheme.slice(ret, end=width - 1) + '\u2026'
                else:
                    ret = grapheme.slice(ret, end=width)
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def __str__(self):
        return '{} {} {} ({})'.format(
            strops.bibser2bibstr(self.__store['no'], self.__store['series']),
            self['first'].title(), self['last'].upper(), self['org'].upper())

    def __repr__(self):
        return 'rider({})'.format(self.__store)

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        """Use a default value id, but don't save it."""
        key = colkey(key)
        if key in self.__store:
            return self.__store[key]
        elif key in _RIDER_DEFAULTS:
            return _RIDER_DEFAULTS[key]
        else:
            return ''

    def __setitem__(self, key, value):
        key = colkey(key)
        self.__store[key] = value
        if key in ['first', 'last', 'org']:
            self.__strcache = {}
        self.__notify(self.get_id())

    def __delitem__(self, key):
        key = colkey(key)
        del (self.__store[key])
        if key in ['fir', 'las', 'org']:
            self.__strcache = {}
        self.__notify(self.get_id())

    def __iter__(self):
        return iter(self.__store.keys())

    def iterkeys(self):
        return iter(self.__store.keys())

    def __contains__(self, item):
        key = colkey(item)
        return key in self.__store

    def __def_notify(self, data=None):
        pass


class riderdb():
    """Rider database."""

    def clear(self):
        """Clear rider model."""
        self.__store = {}
        _log.debug('Rider model cleared')
        self.__notify(None)

    def add_rider(self, newrider, notify=True):
        """Append newrider to model."""
        rid = newrider.get_id()
        if rid in self.__store:
            _log.warning('Duplicate rider entry: %r', rid)
            rid = (newrider['no'], '-'.join(('dupe', strops.randstr())))
        newrider.set_notify(self.__notify)
        self.__store[rid] = newrider
        if notify:
            self.__notify(rid)
        return rid

    def __loadrow(self, r, colspec):
        nr = rider()
        for i in range(0, len(colspec)):
            if len(r) > i:  # column data in row
                val = cellnorm(r[i])
                key = colspec[i]
                nr[key] = val
        if nr['no']:
            if colkey(nr['no']) in _RIDER_COLUMNS:
                _log.debug('Ignore column header: %r', r)
                return
        else:
            _log.warning('Rider without number: %r', nr)
        self.add_rider(nr, notify=False)

    def load(self, csvfile=None):
        """Load riders from supplied CSV file."""
        if not os.path.isfile(csvfile):
            _log.debug('Riders file %r not found', csvfile)
            return
        _log.debug('Loading riders from %r', csvfile)
        with open(csvfile, 'r', encoding='utf-8') as f:
            cr = csv.reader(f)
            incols = None  # no header
            for r in cr:
                if len(r) > 0:  # got a data row
                    if incols is not None:  # already got col header
                        self.__loadrow(r, incols)
                    else:
                        # determine input column structure
                        if colkey(r[0]) in _RIDER_COLUMNS:
                            incols = []
                            for col in r:
                                incols.append(colkey(col))
                        else:
                            incols = _DEFAULT_COLUMN_ORDER  # assume full
                            self.__loadrow(r, incols)
        self.__notify(None)

    def save(self, csvfile=None, columns=None):
        """Save current model content to CSV file."""
        _log.debug('Saving riders to %r', csvfile)
        cats = []
        if columns is None:
            columns = self.include_cols
        with metarace.savefile(csvfile) as f:
            cr = csv.writer(f, quoting=csv.QUOTE_ALL)
            cr.writerow(get_header(columns))
            for r in self.__store.values():
                if r['series'] == 'cat':
                    cats.append(r)
                else:
                    cr.writerow(r.get_row(columns))

            if cats:
                cr.writerow(get_header(columns, _CATEGORY_COLUMNS))
                for r in cats:
                    cr.writerow(r.get_row(columns))

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        return self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = value  # no change to key
        self.__notify(key)

    def __delitem__(self, key):
        del (self.__store[key])
        self.__notify(None)

    def __iter__(self):
        return self.__store.__iter__()

    def get_cat(self, catid):
        """Return a handle to a category entry, or None"""
        return self.get_rider(catid, 'cat')

    def next_rider(self, riderno, series=None):
        """Try to return next rider from series"""
        ret = None
        curid = self.get_id(riderno, series)
        if curid is not None:
            curSeries = curid[1]
            aux = []
            count = 0
            for r in self.__store:
                if curSeries == r[1]:
                    aux.append((strops.bibstr_key(r[0]), count, r))
                    count += 1
            aux.sort()
            i = 0
            rv = None
            while i < len(aux) - 1:
                if aux[i][2] == curid:
                    rv = i + 1
                    break
                i += 1
            if rv is not None:
                ret = aux[rv][2]
        return ret

    def add_empty(self, riderno, series=None):
        """Add a new entry for this rider and series"""
        ret = None
        if series is None:
            riderno, series = strops.bibstr2bibser(riderno)
        nr = rider(no=riderno, series=series)
        return self.add_rider(nr)

    def get_rider(self, riderno, series=None):
        """If rider exists, return handle else return None"""
        ret = None
        rkey = self.get_id(riderno, series)
        if rkey in self.__store:
            ret = self.__store[rkey]
        return ret

    def get_id(self, riderno, series=None):
        """If rider exists, return id else return None"""
        ret = None
        if series is None:
            riderno, series = strops.bibstr2bibser(riderno)
        rkey = (riderno.lower(), series.lower())
        if rkey in self.__store:
            ret = rkey
        return ret

    def iterkeys(self):
        return self.__store.iterkeys()

    def __contains__(self, item):
        return item in self.__store

    def set_notify(self, cb=None):
        """Set the data change notification callback."""
        if cb is None:
            cb = self.__defnotify
        self.__notify = cb
        for ev in self:
            ev.set_notify(cb)

    def __def_notify(self, data=None):
        """Handle changes in db."""
        _log.debug('Notify: %r', data)

    def __init__(self, racetypes=None):
        """Constructor for the event db."""
        self.__store = {}
        self.__notify = self.__def_notify
        self.include_cols = _DEFAULT_COLUMN_ORDER



# SPDX-License-Identifier: MIT
"""export result files in a thread"""

import threading
import subprocess
import logging
import os

import metarace

MIRROR_CMD = 'echo'  # Command/Argument defaults
MIRROR_ARGS = ['dummy', 'srcdir={srcdir}', 'dstdir={dstdir}']
MIRROR_TIMEOUT = 30
LOG = logging.getLogger('metarace.export')
LOG.setLevel(logging.DEBUG)


class mirror(threading.Thread):
    """Mirror thread object class."""

    def __init__(self,
                 callback=None,
                 callbackdata=None,
                 localpath='.',
                 remotepath=None,
                 mirrorcmd=None,
                 arguments=None,
                 data=None):
        """Construct mirror thread object."""
        threading.Thread.__init__(self, daemon=True)
        self.__cb = None
        if callback is not None:
            self.__cb = callback
        self.__cbdata = None
        if callbackdata is not None:
            self.__cbdata = callbackdata
        self.__localpath = localpath
        self.__remotepath = ''
        if remotepath is not None:
            self.__remotepath = remotepath

        # config starts with module defaults
        self.__mirrorcmd = MIRROR_CMD
        self.__arguments = MIRROR_ARGS
        self.__timeout = MIRROR_TIMEOUT

        # then overwrite from sysconf - if present
        if metarace.sysconf.has_section('export'):
            if metarace.sysconf.has_option('export', 'command'):
                self.__mirrorcmd = metarace.sysconf.get('export', 'command')
            if metarace.sysconf.has_option('export', 'arguments'):
                self.__arguments = metarace.sysconf.get('export', 'arguments')
            if metarace.sysconf.has_option('export', 'timeout'):
                self.__timeout = metarace.sysconf.get('export', 'timeout')

        # and then finally allow override in object creation
        if mirrorcmd:
            self.__mirrorcmd = mirrorcmd
        if arguments:
            self.__arguments = arguments

        self.__data = data

    def set_remotepath(self, pathstr):
        """Set or clear the remote path value."""
        self.__remotepath = pathstr

    def set_localpath(self, pathstr):
        """Set or clear the local path value."""
        self.__localpath = pathstr

    def set_cb(self, func=None, cbdata=None):
        """Set or clear the event callback."""
        if func is not None:
            self.__cb = func
            self.__cbdata = cbdata
        else:
            self.__cb = None
            self.__cbdata = None

    def run(self):
        """Called via threading.Thread.start()."""
        running = True
        LOG.debug('Starting')
        ret = None
        try:
            # format errors in arguments caught as exception
            arglist = [
                a.format(srcdir=self.__localpath,
                         dstdir=self.__remotepath,
                         command=self.__mirrorcmd,
                         data=self.__data) for a in self.__arguments
            ]
            arglist.insert(0, self.__mirrorcmd)

            LOG.debug('Calling subprocess: %r', arglist)
            ret = subprocess.run(arglist,
                                 timeout=self.__timeout,
                                 check=True,
                                 capture_output=True)
            if self.__cb is not None:
                self.__cb(ret, self.__cbdata)
        except subprocess.CalledProcessError as e:
            LOG.error('%r returned %r: %s', self.__mirrorcmd, e.returncode,
                      e.stderr.decode('utf8', 'ignore'))
        except Exception as e:
            LOG.error('%s: %s', e.__class__.__name__, e)
        LOG.debug('Complete: Returned %r', ret)



# SPDX-License-Identifier: MIT
"""A collection of tools for preparing cycle race results."""

import os
import logging
import fcntl
import errno
from tempfile import NamedTemporaryFile
from shutil import copyfile
from importlib_resources import files, as_file
from metarace import jsonconfig

VERSION = '2.0.3'
DATA_PATH = os.path.realpath(
    os.path.expanduser(os.path.join('~', 'Documents', 'metarace')))
DEFAULTS_PATH = os.path.join(DATA_PATH, 'default')
RESOURCE_PKG = 'metarace.data'
LOGO = 'metarace_icon.svg'
SYSCONF = 'metarace.json'
PDF_TEMPLATE = 'pdf_template.json'
HTML_TEMPLATE = 'html_template.json'
LOGFILEFORMAT = '%(asctime)s %(levelname)s:%(name)s: %(message)s'
LOGFORMAT = '%(levelname)s:%(name)s: %(message)s'
LOGLEVEL = logging.DEBUG  # default console log level
sysconf = jsonconfig.config()  # system-defaults, populated by init() method
LOG = logging.getLogger('metarace')
LOG.setLevel(logging.DEBUG)


def init():
    """Shared metarace program initialisation."""
    copyconf = mk_data_path()

    # Set global logging options
    logging._srcfile = None
    logging.logThreads = 0
    logging.logProcesses = 0

    # read in a system configuration
    conffile = default_file(SYSCONF)
    try:
        if os.path.exists(conffile):
            LOG.debug('Loading system defaults from %r', conffile)
            with open(conffile) as f:
                sysconf.read(f)
            # don't copy path-specific config into defaults
            copyconf = False
        else:
            LOG.info('System defaults not present, using package defaults')
            ref = files(RESOURCE_PKG).joinpath(SYSCONF)
            with ref.open('r', encoding='utf-8') as f:
                sysconf.read(f)
            copyconf = True
    except Exception as e:
        LOG.error('%s reading system config: %s', e.__class__.__name__, e)

    # if required, create a new system default file
    if copyconf:
        LOG.info('Creating default system config %s', SYSCONF)
        with savefile(os.path.join(DEFAULTS_PATH, SYSCONF)) as f:
            sysconf.write(f)


def mk_data_path():
    """Create a shared data path if it does not yet exist."""
    ret = False
    if not os.path.exists(DATA_PATH):
        LOG.info('Creating data directory: %r', DATA_PATH)
        os.makedirs(DATA_PATH)
    if not os.path.exists(DEFAULTS_PATH):
        LOG.info('Creating system defaults directory: %r', DEFAULTS_PATH)
        os.makedirs(DEFAULTS_PATH)
        ret = True  # flag copy of config back to defaults path
    return ret


def config_path(configpath=None):
    """Clean and check argument for a writeable meet configuration path."""
    ret = None
    if configpath is not None:
        # sanitise into expected config path
        ret = configpath
        if not os.path.isdir(ret):
            ret = os.path.dirname(ret)  # assume dangling path contains file
        ret = os.path.realpath(ret)
        LOG.debug('Checking for meet %r using %r', configpath, ret)
        # then check if the path exists
        if not os.path.exists(ret):
            try:
                LOG.info('Creating meet folder %r', ret)
                os.makedirs(ret)
            except Exception as e:
                LOG.error('Unable to create folder %r: %s', ret, e)
                ret = None
        # check the path is writable
        if ret is not None:
            try:
                LOG.debug('Checking folder %r for write access', ret)
                with NamedTemporaryFile(dir=ret, prefix='.chkwrite_') as f:
                    pass
            except Exception as e:
                LOG.error('Unable to access meet folder %r: %s', ret, e)
                ret = None
    return ret


def default_file(filename=''):
    """Return a path to the named file.

    Path components are stripped, then the the following locations
    are checked in order to find the first instance of filename:
        - current working directory
        - DEFAULTS_PATH
    """
    basefile = os.path.basename(filename)
    if basefile in ['..', '.', '', None]:
        LOG.debug('Invalid filename %r ignored', filename)
        return None
    ret = basefile
    if os.path.exists(basefile):
        pass
    else:
        try:
            check = os.path.join(DEFAULTS_PATH, basefile)
            os.stat(check)
            ret = check
        except Exception as e:
            LOG.debug('%s: %s', e.__class__.__name__, e)
    return ret


def resource_text(name=''):
    """Return a string from the contents of the named resource."""
    basefile = os.path.basename(name)
    if basefile in ['..', '.', '', None]:
        raise FileNotFoundError('Invalid resource name: ' + repr(name))
    t = files(RESOURCE_PKG).joinpath(basefile)
    LOG.debug('Fetching %r from resource %r', basefile, t)
    if t is not None and t.is_file():
        return t.read_text(encoding='utf-8')
    else:
        raise FileNotFoundError('Named resource not found: ' + repr(name))


def resource_file(name=''):
    """Return a temporary filename context manager for a named resource.

    Note: This returns a context manager for a (potentially) temporary
          filename on the filesystem, it must be used in a with statement
          eg:

          with resource_file('resource.svg') as r:
              Gtk.Image.new_from_file(r)
    """

    basefile = os.path.basename(name)
    if basefile in ['..', '.', '', None]:
        raise FileNotFoundError('Invalid resource name: ' + repr(name))
    t = files(RESOURCE_PKG).joinpath(basefile)
    LOG.debug('Fetching %r from resource %r', basefile, t)
    if t is not None and t.is_file():
        return as_file(t)
    else:
        raise FileNotFoundError('Named resource not found: ' + repr(name))


class savefile(object):
    """Tempfile-backed save file contextmanager.

       Creates a temporary file with the desired mode and encoding
       and returns a context manager and writable file handle.

       On close, the temp file is atomically moved to the provided
       filename (if possible).

       Note: This function will log a warning if the file could not be
       moved atomically.
    """

    def __init__(self, filename, mode='t', encoding='utf-8', tempdir='.'):
        self.__sfile = filename
        self.__path = tempdir
        if mode == 'b':
            encoding = None
        self.__tfile = NamedTemporaryFile(mode='w' + mode,
                                          suffix='.tmp',
                                          prefix='sav_',
                                          dir=self.__path,
                                          encoding=encoding,
                                          delete=False)

    def __enter__(self):
        return self.__tfile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__tfile.close()
        if exc_type is not None:
            return False  # raise exception
        # otherwise, file is saved ok in temp file
        os.chmod(self.__tfile.name, 0o644)
        try:
            os.rename(self.__tfile.name, self.__sfile)
            #LOG.debug('os.rename: %r,%r', self.__tfile.name, self.__sfile)
        except OSError as e:
            LOG.debug('os.rename failed: %s', e)
            copyfile(self.__tfile.name, self.__sfile)
            LOG.warn('Un-safely moved file: %r', self.__sfile)
            os.unlink(self.__tfile.name)
        return True


def lockpath(configpath):
    """Open an advisory lock file in the meet config path."""
    lf = None
    lfn = os.path.join(configpath, '.lock')
    try:
        lf = open(lfn, 'a+b')
        fcntl.flock(lf, fcntl.LOCK_EX | fcntl.LOCK_NB)
        LOG.debug('Config lock %r acquired', lfn)
    except Exception as e:
        if lf is not None:
            lf.close()
            lf = None
        LOG.error('Unable to acquire config lock %r: %s', lfn, e)
    return lf


def unlockpath(configpath, lockfile):
    """Release advisory lock and remove lock file."""
    lfn = os.path.join(configpath, '.lock')
    os.unlink(lfn)
    lockfile.close()
    LOG.debug('Config lock %r released', lfn)
    return None


LICENSETEXT = """
MIT License

Copyright (c) 2012-2022 Nathan Fraser and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



# SPDX-License-Identifier: MIT
"""Time of Day types and functions.

Time of Day (tod) records are used to compute net times,
and to communicate context of timing events. Each tod object
includes the following properties:

	timeval	decimal number of seconds, to 6 places
	index	optional serial number or marker (from timing device)
	chan	optional timing channel number or indicator
	refid	optional transponder id or rider identifier
	source	optional source id of timing device that generated the tod

Two specific types are provided:

	tod	Time of Day, strictly >= 0 and less than 24 hours,
		arithmetic is mod 24 hours.
	agg	Aggregate time, may be greater than 24 hours,
		negative values permitted.

Supported arithmetic operations:

		Y:	tod	agg	int	decimal
	tod - Y		yes	no	no	no	*
	tod + Y		yes	no	no	no	*
	agg - Y		yes	yes	yes	yes
	agg + Y		yes	yes	yes	yes

	* Result is mod 24 hours

"""

import decimal
import re
import logging
from datetime import datetime
from dateutil.parser import isoparse, parse as dateparse

# module log object
_log = logging.getLogger('metarace.tod')
_log.setLevel(logging.DEBUG)

# Formatting and truncation constants
QUANT_6PLACES = decimal.Decimal('0.000001')
QUANT_5PLACES = decimal.Decimal('0.00001')
QUANT_4PLACES = decimal.Decimal('0.0001')
QUANT_3PLACES = decimal.Decimal('0.001')
QUANT_2PLACES = decimal.Decimal('0.01')
QUANT_1PLACE = decimal.Decimal('0.1')
QUANT_0PLACES = decimal.Decimal('1')
QUANT = [
    QUANT_0PLACES, QUANT_1PLACE, QUANT_2PLACES, QUANT_3PLACES, QUANT_4PLACES,
    QUANT_5PLACES, QUANT_6PLACES
]
QUANT_FW = [2, 4, 5, 6, 7, 8, 9]
QUANT_TWID = [8, 10, 11, 12, 13, 14, 15]
QUANT_PAD = ['     ', '   ', '  ', ' ', '', '', '']
QUANT_OPAD = ['    ', '  ', ' ', '', '', '', '']
MILL = decimal.Decimal(1000000)

# default rounding is toward zero
ROUNDING = decimal.ROUND_DOWN


def now(index='', chan='CLK', refid='', source='host'):
    """Return a tod set to the current local time."""
    return tod(_now2dec(), index, chan, refid, source)


def fromdate(timestr=''):
    """Try to parse the provided str as a local datetime"""
    rd = None
    rt = None
    try:
        # retrieve date and time for the current timezone
        d = dateparse(timestr).astimezone()
        rd = d.date().isoformat()
        tv = 3600 * d.hour + 60 * d.minute + d.second + d.microsecond / MILL
        rt = mktod(tv)
        rt.source = timestr
    except Exception as e:
        _log.debug('fromdate() %s: %s', e.__class__.__name__, e)
    return (rd, rt)


def fromiso(timestr=''):
    """Retrieve date and TOD from 8601 date and time of day string"""
    rd = None
    rt = None
    try:
        # retrieve date and time for the current timezone
        d = isoparse(timestr).astimezone()
        rd = d.date().isoformat()
        tv = 3600 * d.hour + 60 * d.minute + d.second + d.microsecond / MILL
        rt = mktod(tv)
        rt.source = timestr
    except Exception as e:
        _log.debug('fromiso() %s: %s', e.__class__.__name__, e)
    return (rd, rt)


def fromqc(timestr=''):
    """Retrieve date and TOS from Queclink datestamp"""
    rd = None
    rt = None
    if isinstance(timestr, str) and len(timestr) == 14 and timestr.isdigit():
        stamp = '{}T{}Z'.format(timestr[0:8], timestr[8:14])
        (rd, rt) = fromiso(stamp)
    else:
        _log.debug('fromqc(): Invalid datestamp %r', timestr)
    return (rd, rt)


def mkagg(timeval=''):
    """Return agg for given timeval or None."""
    ret = None
    if timeval is not None and timeval != '':
        try:
            ret = agg(timeval)
        except Exception as e:
            _log.debug('mkagg() %s: %s', e.__class__.__name__, e)
    return ret


def mktod(timeval=''):
    """Return tod for given timeval or None."""
    ret = None
    if timeval is not None and timeval != '':
        try:
            ret = tod(timeval)
        except Exception as e:
            _log.debug('mktod() %s: %s', e.__class__.__name__, e)
    return ret


def _now2dec():
    """Create a decimal timevalue for the current local time."""
    dv = datetime.now()
    ret = (dv.microsecond / MILL).quantize(QUANT_4PLACES)
    ret += 3600 * dv.hour + 60 * dv.minute + dv.second
    return ret


def _dec2hm(dectod=None):
    """Return truncated time string in hours and minutes."""
    strtod = None
    if dectod is not None:
        if dectod >= 3600:  # 'HH:MM'
            strtod = '{0}:{1:02}'.format(
                int(dectod) // 3600, (int(dectod) % 3600) // 60)
        else:  # 'M'
            strtod = '{0}'.format(int(dectod) // 60)
    return strtod


def _dec2ms(dectod=None, places=0, minsep=':'):
    """Return stopwatch type string M:SS[.dcmz]."""
    strtod = None
    if dectod is not None:
        sign = ''
        dv = dectod.quantize(QUANT[places], rounding=ROUNDING)
        if dv.is_signed():
            dv = dv.copy_negate()
            sign = '-'
        # '-M:SS.dcmz'
        strtod = '{0}{1}{2}{3:0{4}}'.format(sign,
                                            int(dv) // 60, minsep, dv % 60,
                                            QUANT_FW[places])
    return strtod


def _dec2str(dectod=None, places=4, zeros=False, hoursep='h', minsep=':'):
    """Return formatted string for given tod decimal value.

    Optional argument 'zeros' will use leading zero chars
    up to 24 hours. eg:

             '00h00:01.2345'   zeros=True
                    '1.2345'   zeros=False
    """
    strtod = None
    if dectod is not None:
        sign = ''
        dv = dectod.quantize(QUANT[places], rounding=ROUNDING)
        if dv.is_signed():
            dv = dv.copy_negate()
            sign = '-'
        if zeros or dv >= 3600:  # '-HhMM:SS.dcmz'
            fmt = '{0}{1}{2}{3:02}{4}{5:0{6}}'
            if zeros:  # '-00h00:0S.dcmz'
                fmt = '{0}{1:02}{2}{3:02}{4}{5:0{6}}'
            strtod = fmt.format(sign,
                                int(dv) // 3600, hoursep,
                                (int(dv) % 3600) // 60, minsep, dv % 60,
                                QUANT_FW[places])
        elif dv >= 60:  # '-M:SS.dcmz'
            strtod = '{0}{1}{2}{3:0{4}}'.format(sign,
                                                int(dv) // 60, minsep, dv % 60,
                                                QUANT_FW[places])
        else:  # '-S.dcmz'
            strtod = '{0}{1}'.format(sign, dv)
    return strtod


def _str2dec(timestr=''):
    """Return decimal for given string.

    Attempt to match against patterns:
    	-HhMM:SS.dcmz		Canonical
    	-H:MM:SS.dcmz		Omega
    	-H:MM'SS"dcmz		Chronelec
    	-H-MM-SS.dcmz		Keypad entry
        PThHmMs.dcmzS		ISO8601 Interval
    """
    dectod = None
    timestr = timestr.strip()  # assumes string
    if timestr == 'now':
        dectod = _now2dec()
    elif timestr.startswith('PT'):
        # interpret as ISO8601 Interval (4.4.1b), but only time flags
        m = re.match(
            r'^PT((\d+([\.\,]\d+)?)H)?((\d+([\.\,]\d+)?)M)?((\d+([\.\,]\d+)?)S)?',
            timestr)
        if m is not None:
            dectod = 0
            if m.group(2):
                dectod += 3600 * decimal.Decimal(m.group(2).replace(',', '.'))
            if m.group(5):
                dectod += 60 * decimal.Decimal(m.group(5).replace(',', '.'))
            if m.group(8):
                dectod += decimal.Decimal(m.group(8).replace(',', '.'))
        else:
            _log.info('_str2dec() Invalid 8601 interval: %r', timestr)
    else:
        m = re.match(
            r'^(-?)(?:(?:(\d+)[h:-])?(\d{1,2})[:\'-])?(\d{1,2}(?:[\.\"]\d+)?)$',
            timestr)
        if m is not None:
            dectod = decimal.Decimal(m.group(4).replace('"', '.'))
            dectod += decimal.Decimal(m.group(3) or 0) * 60
            dectod += decimal.Decimal(m.group(2) or 0) * 3600
            if m.group(1):  # negative sign
                dectod = dectod.copy_negate()
        else:
            dectod = decimal.Decimal(timestr)
            #_log.debug('_str2dec() Decimal conversion %s => %s', timestr,
            #dectod)
    return dectod


def _tv2dec(timeval):
    """Convert the provided value into a decimal timeval for tod/agg."""
    ret = 0
    if isinstance(timeval, decimal.Decimal):
        ret = timeval
    elif isinstance(timeval, str):
        ret = _str2dec(timeval)
    elif isinstance(timeval, tod):
        # Discard context on supplied tod and copy decimal obj
        ret = timeval.timeval
    elif isinstance(timeval, float):
        # Round off float to max tod precision
        ret = decimal.Decimal('{0:0.6f}'.format(timeval))
    else:
        ret = decimal.Decimal(timeval)
    return ret


class tod(object):
    """A class for representing time of day, net time and RFID events."""

    def __init__(self,
                 timeval=0,
                 index='',
                 chan='TOD',
                 refid='',
                 source='host'):
        self.index = index
        self.chan = chan
        self.refid = refid
        self.source = source
        self.timeval = _tv2dec(timeval)
        if self.timeval < 0 or self.timeval >= 86400:
            raise ValueError('Time of day value not in range [0, 86400)')

    def __str__(self):
        """Return a normalised tod string."""
        return str(self.__unicode__())

    def __unicode__(self):
        """Return a normalised tod string."""
        return '{0: >5} {1: <3} {2} {3} {4}'.format(self.index, self.chan,
                                                    self.timestr(4),
                                                    self.refid, self.source)

    def __repr__(self):
        """Return object representation string."""
        return "{5}({0}, {1}, {2}, {3}, {4})".format(repr(self.timeval),
                                                     repr(self.index),
                                                     repr(self.chan),
                                                     repr(self.refid),
                                                     repr(self.source),
                                                     self.__class__.__name__)

    def truncate(self, places=4):
        """Return a new truncated time value."""
        return self.__class__(timeval=self.timeval.quantize(QUANT[places],
                                                            rounding=ROUNDING),
                              chan='TRUNC')

    def as_hours(self, places=0):
        """Return decimal value in hours, truncated to the desired places."""
        return (self.timeval / 3600).quantize(QUANT[places], rounding=ROUNDING)

    def as_minutes(self, places=0):
        """Return decimal value in minutes, truncated to the desired places."""
        return (self.timeval / 60).quantize(QUANT[places], rounding=ROUNDING)

    def as_seconds(self, places=0):
        """Return decimal value in seconds, truncated to the desired places."""
        return self.timeval.quantize(QUANT[places], rounding=ROUNDING)

    def timestr(self, places=4, zeros=False, hoursep='h', minsep=':'):
        """Return time string component of the tod, whitespace padded."""
        return '{0: >{1}}{2}'.format(
            _dec2str(self.timeval, places, zeros, hoursep, minsep),
            QUANT_TWID[places], QUANT_PAD[places])

    def omstr(self, places=3, zeros=False, hoursep=':', minsep=':'):
        """Return a 12 digit 'omega' style time string."""
        if places > 3:
            places = 3  # Hack to clamp to 12 dig
        return '{0: >{1}}{2}'.format(
            _dec2str(self.timeval, places, zeros, hoursep, minsep),
            QUANT_TWID[places], QUANT_OPAD[places])

    def minsec(self, places=0, minsep=':'):
        """Return a stopwatch type string m:ss.[dcmz]."""
        return _dec2ms(self.timeval, places, minsep)

    def meridiem(self, mstr=None, secs=True):
        """Return a 12hour time of day string with meridiem."""
        ret = None
        med = '\u2006am'
        tv = self.timeval
        # unwrap timeval into a single 24hr period
        if tv >= 86400:
            tv = tv % 86400
        elif tv < 0:
            tv = 86400 - (tv.copy_abs() % 86400)

        # determine meridiem and adjust for display
        if tv >= 43200:
            med = '\u2006pm'
        if mstr is not None:
            med = mstr
        tv = tv % 43200
        if tv < 3600:  # 12am/12pm
            tv += 43200
        if secs:
            ret = _dec2str(tv, 0, hoursep=':', minsep=':') + med
        else:
            ret = _dec2hm(tv) + med
        return ret

    def isosecs(self, places=4):
        """Return ISO8601(4.4.1b) Refer : 4.4.3.2 Format with designators."""
        return "PT{}S".format(
            self.timeval.quantize(QUANT[places], rounding=ROUNDING))

    def isostr(self, places=4):
        """Return ISO8601(4.4.1b) Time interval string"""
        return "PT{}S".format(
            _dec2str(self.timeval, places, hoursep='H', minsep='M'))

    def rawtime(self, places=4, zeros=False, hoursep='h', minsep=':'):
        """Return time string of tod as string, without padding."""
        return _dec2str(self.timeval, places, zeros, hoursep, minsep)

    def speedstr(self, dist=200):
        """Return average speed estimate string for the provided distance."""
        if self.timeval == 0:
            return '---.-\u2006km/h'
        return '{0:5.1f}\u2006km/h'.format(3.6 * float(dist) /
                                           float(self.timeval))

    def rawspeed(self, dist=200):
        """Return an average speed estimate string without unit."""
        if self.timeval == 0:
            return '-.-'
        return '{0:0.1f}'.format(3.6 * float(dist) / float(self.timeval))

    def __lt__(self, other):
        if isinstance(other, tod):
            return self.timeval < other.timeval
        else:
            return self.timeval < other

    def __le__(self, other):
        if isinstance(other, tod):
            return self.timeval <= other.timeval
        else:
            return self.timeval <= other

    def __eq__(self, other):
        if isinstance(other, tod):
            return self.timeval == other.timeval
        else:
            return self.timeval == other

    def __ne__(self, other):
        if isinstance(other, tod):
            return self.timeval != other.timeval
        else:
            return self.timeval != other

    def __gt__(self, other):
        if isinstance(other, tod):
            return self.timeval > other.timeval
        else:
            return self.timeval > other

    def __ge__(self, other):
        if isinstance(other, tod):
            return self.timeval >= other.timeval
        else:
            return self.timeval >= other

    def __sub__(self, other):
        """Compute time of day subtraction and return a NET tod object."""
        if type(other) is not tod:  # Subclass must override this method
            return NotImplemented
        if self.timeval >= other.timeval:
            oft = self.timeval - other.timeval
        else:
            oft = 86400 - other.timeval + self.timeval
        return tod(timeval=oft, chan='NET')

    def __add__(self, other):
        """Compute time of day addition and return a new tod object."""
        if type(other) is not tod:  # Subclass must override this method
            return NotImplemented
        return tod(timeval=(self.timeval + other.timeval) % 86400, chan='SUM')

    def __pos__(self):
        """Unary + operation."""
        return self.__class__(self.timeval, chan='POS')

    def __abs__(self):
        """Unary absolute value."""
        return self.__class__(self.timeval.copy_abs(), chan='ABS')


class agg(tod):
    """Aggregate time type."""

    def __init__(self,
                 timeval=0,
                 index='',
                 chan='AGG',
                 refid='',
                 source='host'):
        self.index = index
        self.chan = chan
        self.refid = refid
        self.source = source
        self.timeval = _tv2dec(timeval)

    def __add__(self, other):
        """Compute addition and return aggregate."""
        if isinstance(other, tod):
            return agg(timeval=self.timeval + other.timeval, chan='AGG')
        elif isinstance(other, (int, decimal.Decimal)):
            return agg(timeval=self.timeval + other, chan='AGG')
        else:
            return NotImplemented

    def __sub__(self, other):
        """Compute subtraction and return aggregate."""
        if isinstance(other, tod):
            return agg(timeval=self.timeval - other.timeval, chan='AGG')
        elif isinstance(other, (int, decimal.Decimal)):
            return agg(timeval=self.timeval - other, chan='AGG')
        else:
            return NotImplemented

    def __neg__(self):
        """Unary - operation."""
        return self.__class__(self.timeval.copy_negate(), chan='AGG')


# TOD 'constants'
ZERO = tod()
ONE = tod('1.0')
MINUTE = tod('1:00')
MAX = tod('23h59:59.999990')  # largest val possible for tod
MAXELAP = tod('23h30:00')  # max displayed elapsed time

# Fake times for special cases
# these are unused tods that sort correctly when compared
FAKETIMES = {
    'catch': tod(ZERO, chan='catch'),
    'w/o': tod(ZERO, chan='w/o'),
    'max': tod(MAX, chan='max'),
    'ntr': tod(MAX, chan='ntr'),
    'caught': tod(MAX, chan='caught'),
    'rel': tod(MAX, chan='rel'),
    'abort': tod(MAX, chan='abort'),
    'otl': tod(MAX, chan='otl'),
    'dnf': tod(MAX, chan='dnf'),
    'dns': tod(MAX, chan='dns'),
    'dsq': tod(MAX, chan='dsq'),
}
extra = decimal.Decimal('0.000001')
cof = decimal.Decimal('0.000001')
for c in ['ntr', 'caught', 'rel', 'abort', 'otl', 'dnf', 'dns', 'dsq']:
    FAKETIMES[c].timeval += cof
    cof += extra


class todlist(object):
    """ToD list helper class for managing splits and ranks."""

    def __init__(self, lbl=''):
        self.__label = lbl
        self.__store = []

    def __iter__(self):
        return self.__store.__iter__()

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        return self.__store[key]

    def rank(self, bib, series=''):
        """Return current 0-based rank for given bib."""
        ret = None
        count = 0
        i = 0
        lpri = None
        lsec = None
        for lt in self.__store:
            # scan times for updating ranks
            if lpri is not None:
                if lt[0] != lpri or lt[1] != lsec:
                    i = count
            # if rider matches, break
            if lt[0].refid == bib and lt[0].index == series:
                ret = i
                break
            lpri = lt[0]
            lsec = lt[1]
            count += 1
        return ret
        """Return current 0-based rank for given bib."""
        ret = None
        i = 0
        r = 0
        last = None
        for lt in self.__store:
            if last is not None:
                if lt != last:
                    r = i
            i += 1
            if lt.refid == bib and lt.index == series:
                ret = r
                break
            last = lt
        return ret

    def clear(self):
        self.__store = []
        return 0

    def remove(self, bib, series='', once=False):
        i = 0
        while i < len(self.__store):
            if (self.__store[i][0].refid == bib
                    and self.__store[i][0].index == series):
                del self.__store[i]
                if once:
                    break
            else:
                i += 1
        return i

    def insert(self, pri=None, sec=None, bib=None, series='', prec=3):
        """Insert primary tod and secondary tod into ordered list."""
        ret = None
        trunc = True
        if isinstance(pri, str) and pri in FAKETIMES:
            pri = FAKETIMES[pri]
            trunc = False  # retain precision for primary comparison

        if isinstance(pri, tod):
            if bib is None:
                bib = pri.index
            if sec is None:
                sec = ZERO
            if trunc:
                pri = pri.truncate(prec)
                sec = sec.truncate(prec)
            rt0 = tod(pri, chan=self.__label, refid=bib, index=series)
            rt1 = tod(sec, chan=self.__label, refid=bib, index=series)
            last = None
            i = 0
            found = False
            for lt in self.__store:
                if rt0 < lt[0]:  # primary time is faster, insert ok
                    self.__store.insert(i, [rt0, rt1])
                    found = True
                    break
                elif rt0 == lt[0]:  # primary same, compare on secondary
                    if rt1 < lt[1]:  # secondary time faster - insert ok
                        self.__store.insert(i, [rt0, rt1])
                        found = True
                        break
                i += 1
            if not found:
                self.__store.append([rt0, rt1])
            ret = i
        return ret



# SPDX-License-Identifier: MIT
"""HTML output library.

Cheap and nasty functional primitives for generating loosely
compliant HTML output. Each element primitive returns a single
escaped string.

All elements take a named parameter 'attrs' which is a dict of
key/value attributes. Non-empty elements take a parameter 'elements'
which is a list of child elements or a plain string.
"""

from xml.sax.saxutils import escape, quoteattr
import sys


class element(str):
    """String wrapper for serialised HTML text"""
    pass


def doctype():
    """HTML doctype pseudo-element"""
    return element('<!doctype html>')


def attrlist(attrs):
    """Convert attr dict into escaped attrlist."""
    alist = []
    for a in attrs:
        alist.append(a.lower() + '=' + quoteattr(attrs[a]))
    if len(alist) > 0:
        alist.insert(0, '')
        return ' '.join(alist)
    else:
        return ''


def escapetext(text=''):
    """Return escaped copy of text."""
    return element(escape(text, {'"': '&quot;'}))


def serialise(elements=()):
    """Concatenate element list into an escaped string."""
    if isinstance(elements, str):
        elements = (elements, )
    elist = []
    for j in elements:
        if type(j) is element:
            elist.append(j)
        else:
            elist.append(escapetext(j))
    return '\n'.join(elist)


def comment(elements=(' ', )):
    """Wrap elements with a comment marker"""
    text = serialise(elements).replace('<!--',
                                       '').replace('-->',
                                                   '').replace('--!>',
                                                               '').lstrip('>')
    if text.endswith('<!-'):
        text = text.rstrip('-')
    return element(''.join(('<!--', text, '-->')))


# return a valid but empty html template
def emptypage():
    return '\n'.join((
        doctype(),
        html((
            head((
                meta(attrs={'charset': 'utf-8'}),
                meta(
                    attrs={
                        'name': 'viewport',
                        'content': 'width=device-width, initial-scale=1'
                    }),
                title('__REPORT_TITLE__'),
                link(
                    attrs={
                        'href':
                        'https://cdn.jsdelivr.net/npm/bootstrap@5.2.0-beta1/dist/css/bootstrap.min.css',
                        'integrity':
                        'sha384-0evHe/X+R7YkIZDRvuzKMRqM+OrBnVFBL6DOitfPri4tjfHxaWutUpFmBp4vmVor',
                        'crossorigin': 'anonymous',
                        'rel': 'stylesheet'
                    }),
                link(
                    attrs={
                        'href':
                        'https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.3/font/bootstrap-icons.css',
                        'rel': 'stylesheet'
                    }),
                script(
                    element(
                        'function ud(){null!==document.querySelector("#pgre")&&setTimeout("history.go(0);",55329)}function rl(){return setTimeout("history.go(0);",10),!1}'
                    ), ),
            )),
            body((
                '__REPORT_NAV__',
                div((
                    h1('__REPORT_TITLE__'),
                    '__REPORT_CONTENT__',
                ),
                    attrs={'class': 'container'}),
            ),
                 attrs={'onload': 'ud();'}),
        ), {'lang': 'en'})))


# Declare all the empty types
for empty in ('base', 'link', 'meta', 'hr', 'br', 'wbr', 'source', 'img',
              'embed', 'track', 'area', 'col', 'param', 'hr', 'br', 'img',
              'col'):

    def emptyfunc(attrs={}, tag=empty):
        return element(''.join(('<', tag, attrlist(attrs), '>')))

    setattr(sys.modules[__name__], empty, emptyfunc)


def forminput(attrs={}):
    return element(''.join(('<input', attrlist(attrs), '>')))


# Declare all the non-empty elements (except specials defined above)
for nonempty in (
        'html',
        'head',
        'title',
        'style',
        'body',
        'article',
        'section',
        'nav',
        'aside',
        'h1',
        'h2',
        'h3',
        'h4',
        'h5',
        'h6',
        'hgroup',
        'header',
        'footer',
        'address',
        'p',
        'pre',
        'blockquote',
        'ol',
        'ul',
        'menu',
        'li',
        'dl',
        'dt',
        'dd',
        'figure',
        'figcaption',
        'main',
        'div',
        'a',
        'em',
        'strong',
        'small',
        's',
        'cite',
        'q',
        'dfn',
        'abbr',
        'ruby',
        'rt',
        'rp',
        'data',
        'time',
        'code',
        'var',
        'samp',
        'kbd',
        'sub',
        'sup',
        'i',
        'b',
        'u',
        'mark',
        'bdi',
        'bdo',
        'span',
        'ins',
        'del',
        'picture',
        'object',
        'video',
        'audio',
        'map',
        'math',
        'svg',
        'table',
        'caption',
        'colgroup',
        'tbody',
        'thead',
        'tfoot',
        'tr',
        'td',
        'th',
        'form',
        'label',
        'button',
        'select',
        'datalist',
        'optgroup',
        'option',
        'textarea',
        'output',
        'progress',
        'meter',
        'fieldset',
        'legend',
        'details',
        'summary',
        'dialog',
        'script',
        'slot',
        'canvas',
):

    def nonemptyfunc(elements=(), attrs={}, elem=nonempty):
        return element(''.join(('<', elem, attrlist(attrs), '>',
                                serialise(elements), '</', elem, '>')))

    setattr(sys.modules[__name__], nonempty, nonemptyfunc)
